@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2002-2004 MontaVista Software, Inc.
- * Copyright (c) 2006-2009 Red Hat, Inc.
+ * Copyright (c) 2006-2010 Red Hat, Inc.
  *
  * Author: Steven Dake (sdake@redhat.com)
  * Author: Lon Hohberger (lhh@redhat.com)
@@ -37,6 +37,7 @@
 
 #include <config.h>
 
+#include <stdint.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <string.h>
@@ -59,7 +60,10 @@
 #include <stdlib.h>
 #include <pthread.h>
 #include <limits.h>
+#include <sys/mman.h>
+#include <semaphore.h>
 
+#include <corosync/list.h>
 #include <corosync/engine/logsys.h>
 
 #define YIELD_AFTER_LOG_OPS 10
@@ -115,28 +119,15 @@ struct syslog_names facilitynames[] =
 	{ NULL, -1 }
 };
 
-
-/*
- * These are not static so they can be read from the core file
- */
-int *flt_data;
-
-unsigned int flt_data_size;
-
-#define COMBINE_BUFFER_SIZE 2048
-
-/* values for logsys_logger init_status */
-#define LOGSYS_LOGGER_INIT_DONE		0
-#define LOGSYS_LOGGER_NEEDS_INIT	1
-
-static int logsys_system_needs_init = LOGSYS_LOGGER_NEEDS_INIT;
-
-static int logsys_sched_param_queued = 0;
-static int logsys_sched_policy;
-static struct sched_param logsys_sched_param;
-
-static int logsys_after_log_ops_yield = 10;
-
+struct record {
+	unsigned int rec_ident;
+	const char *file_name;
+	const char *function_name;
+	int file_line;
+	char *buffer;
+	struct list_head list;
+};
+	
 /*
  * need unlogical order to preserve 64bit alignment
  */
@@ -153,9 +144,35 @@ struct logsys_logger {
 						   for subsystems */
 };
 
+
 /*
- * operating global variables
+ * These are not static so they can be read from the core file
  */
+int *flt_data;
+
+uint32_t flt_head;
+
+uint32_t flt_tail;
+
+unsigned int flt_data_size;
+
+#define COMBINE_BUFFER_SIZE 2048
+
+/* values for logsys_logger init_status */
+#define LOGSYS_LOGGER_INIT_DONE		0
+#define LOGSYS_LOGGER_NEEDS_INIT	1
+
+static int logsys_system_needs_init = LOGSYS_LOGGER_NEEDS_INIT;
+
+static int logsys_memory_used = 0;
+
+static int logsys_sched_param_queued = 0;
+
+static int logsys_sched_policy;
+
+static struct sched_param logsys_sched_param;
+
+static int logsys_after_log_ops_yield = 10;
 
 static struct logsys_logger logsys_loggers[LOGSYS_MAX_SUBSYS_COUNT + 1];
 
@@ -169,156 +186,164 @@ static unsigned int records_written = 1;
 
 static pthread_t logsys_thread_id;
 
-static pthread_cond_t logsys_cond;
+static sem_t logsys_thread_start;
 
-static pthread_mutex_t logsys_cond_mutex;
+static sem_t logsys_print_finished;
 
 #if defined(HAVE_PTHREAD_SPIN_LOCK)
-static pthread_spinlock_t logsys_idx_spinlock;
+static pthread_spinlock_t logsys_flt_spinlock;
 #else
-static pthread_mutex_t logsys_idx_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t logsys_flt_mutex = PTHREAD_MUTEX_INITIALIZER;
 #endif
 
-static unsigned int log_rec_idx;
+#if defined(HAVE_PTHREAD_SPIN_LOCK)
+static pthread_spinlock_t logsys_wthread_spinlock;
+#else
+static pthread_mutex_t logsys_wthread_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
 
 static int logsys_buffer_full = 0;
 
 static char *format_buffer=NULL;
 
-static int log_requests_pending = 0;
-
-static int log_requests_lost = 0;
+static int logsys_dropped_messages = 0;
 
 void *logsys_rec_end;
 
-#define FDHEAD_INDEX	(flt_data_size)
-
-#define FDTAIL_INDEX 	(flt_data_size + 1)
+static DECLARE_LIST_INIT(logsys_print_finished_records);
 
 #define FDMAX_ARGS	64
 
+#define CIRCULAR_BUFFER_WRITE_SIZE	64
+
 /* forward declarations */
 static void logsys_close_logfile(int subsysid);
 
-#ifdef LOGSYS_DEBUG
-static char *decode_mode(int subsysid, char *buf, size_t buflen)
-{
-	memset(buf, 0, buflen);
-
-	if (logsys_loggers[subsysid].mode & LOGSYS_MODE_OUTPUT_FILE)
-		snprintf(buf+strlen(buf), buflen, "FILE,");
-
-	if (logsys_loggers[subsysid].mode & LOGSYS_MODE_OUTPUT_STDERR)
-		snprintf(buf+strlen(buf), buflen, "STDERR,");
-
-	if (logsys_loggers[subsysid].mode & LOGSYS_MODE_OUTPUT_SYSLOG)
-		snprintf(buf+strlen(buf), buflen, "SYSLOG,");
-
-	if (subsysid == LOGSYS_MAX_SUBSYS_COUNT) {
-		if (logsys_loggers[subsysid].mode & LOGSYS_MODE_FORK)
-			snprintf(buf+strlen(buf), buflen, "FORK,");
-
-		if (logsys_loggers[subsysid].mode & LOGSYS_MODE_THREADED)
-			snprintf(buf+strlen(buf), buflen, "THREADED,");
-	}
-
-	memset(buf+strlen(buf)-1,0,1);
-
-	return buf;
-}
-
-static const char *decode_debug(int subsysid)
-{
-	if (logsys_loggers[subsysid].debug)
-		return "on";
-
-	return "off";
-}
-
-static const char *decode_status(int subsysid)
-{
-	if (!logsys_loggers[subsysid].init_status)
-		return "INIT_DONE";
-
-	return "NEEDS_INIT";
-}
-
-static void dump_subsys_config(int subsysid)
-{
-	char modebuf[1024];
-
-	fprintf(stderr,
-		"ID: %d\n"
-		"subsys: %s\n"
-		"logfile: %s\n"
-		"logfile_fp: %p\n"
-		"mode: %s\n"
-		"debug: %s\n"
-		"syslog_fac: %s\n"
-		"syslog_pri: %s\n"
-		"logfile_pri: %s\n"
-		"init_status: %s\n",
-		subsysid,
-		logsys_loggers[subsysid].subsys,
-		logsys_loggers[subsysid].logfile,
-		logsys_loggers[subsysid].logfile_fp,
-		decode_mode(subsysid, modebuf, sizeof(modebuf)),
-		decode_debug(subsysid),
-		logsys_facility_name_get(logsys_loggers[subsysid].syslog_facility),
-		logsys_priority_name_get(logsys_loggers[subsysid].syslog_priority),
-		logsys_priority_name_get(logsys_loggers[subsysid].logfile_priority),
-		decode_status(subsysid));
-}
-
-static void dump_full_config(void)
+static uint32_t circular_memory_map (void **buf, size_t bytes)
 {
+	void *addr_orig;
+	void *addr;
+	int fd;
+	int res;
+	const char *file = "fdata-XXXXXX";
+	char path[PATH_MAX];
+	char buffer[CIRCULAR_BUFFER_WRITE_SIZE];
 	int i;
+	int written;
+	int error_return = 0;
 
-	for (i = 0; i <= LOGSYS_MAX_SUBSYS_COUNT; i++) {
-		if (strlen(logsys_loggers[i].subsys) > 0)
-			dump_subsys_config(i);
+	snprintf (path, PATH_MAX, "/dev/shm/%s", file);
+
+	fd = mkstemp (path);
+	if (fd == -1) {
+		snprintf (path, PATH_MAX, LOCALSTATEDIR "/run/%s", file);
+		fd = mkstemp (path);
+		if (fd == -1) {
+			error_return = -1;
+			goto error_exit;
+		}
 	}
-}
+
+	/*
+	 * ftruncate doesn't return ENOSPC
+	 * have to use write to determine if shared memory is actually available
+	 */
+	res = ftruncate (fd, 0);
+	if (res == -1) {
+		error_return = -1;
+		goto unlink_exit;
+	}
+	memset (buffer, 0, sizeof (buffer));
+	for (i = 0; i < (bytes / CIRCULAR_BUFFER_WRITE_SIZE); i++) {
+retry_write:
+		written = write (fd, buffer, CIRCULAR_BUFFER_WRITE_SIZE);
+		if (written == -1 && errno == EINTR) {
+			goto retry_write;
+		}
+		if (written != 64) {
+			error_return = -1;
+			goto unlink_exit;
+		}
+	}
+	
+	addr_orig = mmap (NULL, bytes << 1, PROT_NONE,
+		MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
+	if (addr_orig == MAP_FAILED) {
+		error_return = -1;
+		goto unlink_exit;
+	}
+
+	addr = mmap (addr_orig, bytes, PROT_READ | PROT_WRITE,
+		MAP_SHARED | MAP_FIXED, fd, 0);
+	if (addr != addr_orig) {
+		error_return = -1;
+		goto mmap_exit;
+	}
+	#ifdef COROSYNC_BSD
+		madvise(addr_orig, bytes, MADV_NOSYNC);
+	#endif
+
+	addr = mmap (((char *)addr_orig) + bytes,
+		  bytes, PROT_READ | PROT_WRITE,
+		  MAP_SHARED | MAP_FIXED, fd, 0);
+	if ((char *)addr != (char *)((char *)addr_orig + bytes)) {
+		error_return = -1;
+		goto mmap_exit;
+	}
+#ifdef COROSYNC_BSD
+	madvise(((char *)addr_orig) + bytes, bytes, MADV_NOSYNC);
 #endif
 
-/*
- * Helpers for _logsys_log_rec functionality
- */
-static inline void my_memcpy_32bit (int *dest, const int *src, unsigned int words)
-{
-	unsigned int word_idx;
-	for (word_idx = 0; word_idx < words; word_idx++) {
-		dest[word_idx] = src[word_idx];
-	}
-}
+	*buf = addr_orig;
+	error_return = 0;
+	goto unlink_exit;
 
-static inline void my_memcpy_8bit (char *dest, const char *src,
-				   unsigned int bytes)
-{
-	unsigned int byte_idx;
-
-	for (byte_idx = 0; byte_idx < bytes; byte_idx++) {
-		dest[byte_idx] = src[byte_idx];
-	}
+mmap_exit:
+	munmap (addr_orig, bytes << 1);
+unlink_exit:
+	unlink (path);
+	close (fd);
+error_exit:
+	return (error_return);
 }
 
 #if defined(HAVE_PTHREAD_SPIN_LOCK)
-static void logsys_lock (void)
+static void logsys_flt_lock (void)
 {
-	pthread_spin_lock (&logsys_idx_spinlock);
+	pthread_spin_lock (&logsys_flt_spinlock);
 }
-static void logsys_unlock (void)
+static void logsys_flt_unlock (void)
 {
-	pthread_spin_unlock (&logsys_idx_spinlock);
+	pthread_spin_unlock (&logsys_flt_spinlock);
 }
 #else
-static void logsys_lock (void)
+static void logsys_flt_lock (void)
 {
-	pthread_mutex_lock (&logsys_idx_mutex);
+	pthread_mutex_lock (&logsys_flt_mutex);
 }
-static void logsys_unlock (void)
+static void logsys_flt_unlock (void)
 {
-	pthread_mutex_unlock (&logsys_idx_mutex);
+	pthread_mutex_unlock (&logsys_flt_mutex);
+}
+#endif
+
+#if defined(HAVE_PTHREAD_SPIN_LOCK)
+static void logsys_wthread_lock (void)
+{
+	pthread_spin_lock (&logsys_wthread_spinlock);
+}
+static void logsys_wthread_unlock (void)
+{
+	pthread_spin_unlock (&logsys_wthread_spinlock);
+}
+#else
+static void logsys_wthread_lock (void)
+{
+	pthread_mutex_lock (&logsys_wthread_mutex);
+}
+static void logsys_wthread_unlock (void)
+{
+	pthread_mutex_unlock (&logsys_wthread_mutex);
 }
 #endif
 
@@ -338,13 +363,12 @@ static inline void records_reclaim (unsigned int idx, unsigned int words)
 		return;
 	}
 
-	logsys_lock();
-	if (flt_data[FDTAIL_INDEX] > flt_data[FDHEAD_INDEX]) {
-		if (idx + words >= flt_data[FDTAIL_INDEX]) {
+	if (flt_tail > flt_head) {
+		if (idx + words >= flt_tail) {
 			should_reclaim = 1;
 		}
 	} else {
-		if ((idx + words) >= (flt_data[FDTAIL_INDEX] + flt_data_size)) {
+		if ((idx + words) >= (flt_tail + flt_data_size)) {
 			should_reclaim = 1;
 		}
 	}
@@ -356,18 +380,13 @@ static inline void records_reclaim (unsigned int idx, unsigned int words)
 		do {
 			unsigned int old_tail;
 
-			words_needed -= flt_data[flt_data[FDTAIL_INDEX]];
-			old_tail = flt_data[FDTAIL_INDEX];
-			flt_data[FDTAIL_INDEX] =
-				(flt_data[FDTAIL_INDEX] +
-				flt_data[flt_data[FDTAIL_INDEX]]) % (flt_data_size);
-			if (log_rec_idx == old_tail) {
-				log_requests_lost += 1;
-				log_rec_idx = flt_data[FDTAIL_INDEX];
-			}
+			words_needed -= flt_data[flt_tail];
+			old_tail = flt_tail;
+			flt_tail =
+				(flt_tail +
+				flt_data[flt_tail]) % (flt_data_size);
 		} while (words_needed > 0);
 	}
-	logsys_unlock();
 }
 
 #define idx_word_step(idx)						\
@@ -631,159 +650,108 @@ static void log_printf_to_logs (
 	}
 }
 
-static void record_print (const char *buf)
+static void log_printf_to_logs_wthread (
+	unsigned int rec_ident,
+	const char *file_name,
+	const char *function_name,
+	int file_line,
+	const char *buffer)
 {
-	const int *buf_uint32t = (const int *)buf;
-	unsigned int rec_size = buf_uint32t[0];
-	unsigned int rec_ident = buf_uint32t[1];
-	unsigned int file_line = buf_uint32t[2];
-	unsigned int i;
-	unsigned int words_processed;
-	unsigned int arg_size_idx;
-	const void *arguments[FDMAX_ARGS];
-	unsigned int arg_count;
+	struct record *rec;
+	uint32_t length;
 
-	arg_size_idx = 4;
-	words_processed = 4;
-	arg_count = 0;
-
-	for (i = 0; words_processed < rec_size; i++) {
-		arguments[arg_count++] = &buf_uint32t[arg_size_idx + 1];
-		arg_size_idx += buf_uint32t[arg_size_idx] + 1;
-		words_processed += buf_uint32t[arg_size_idx] + 1;
-	}
-
-	/*
-	 * (char *)arguments[0] -> subsystem
-	 * (char *)arguments[1] -> file_name
-	 * (char *)arguments[2] -> function_name
-	 * (char *)arguments[3] -> message
-	 */
-
-	log_printf_to_logs (
-		rec_ident,
-		(char *)arguments[1],
-		(char *)arguments[2],
-		file_line,
-		(char *)arguments[3]);
-}
-
-static int record_read (char *buf, int rec_idx, int *log_msg) {
-	unsigned int rec_size;
-	unsigned int rec_ident;
-	int firstcopy, secondcopy;
-
-	rec_size = flt_data[rec_idx];
-	rec_ident = flt_data[(rec_idx + 1) % flt_data_size];
-
-	/*
-	 * Not a log record
-	 */
-	if (LOGSYS_DECODE_RECID(rec_ident) != LOGSYS_RECID_LOG) {
-		*log_msg = 0;
-        	return ((rec_idx + rec_size) % flt_data_size);
-	}
-
-	/*
-	 * A log record
-	 */
-	*log_msg = 1;
-
-        firstcopy = rec_size;
-        secondcopy = 0;
-        if (firstcopy + rec_idx > flt_data_size) {
-                firstcopy = flt_data_size - rec_idx;
-                secondcopy -= firstcopy - rec_size;
-        }
-        memcpy (&buf[0], &flt_data[rec_idx], firstcopy << 2);
-        if (secondcopy) {
-                memcpy (&buf[(firstcopy << 2)], &flt_data[0], secondcopy << 2);
-        }
-        return ((rec_idx + rec_size) % flt_data_size);
-}
-
-static inline void wthread_signal (void)
-{
-	if (wthread_active == 0) {
+	rec = malloc (sizeof (struct record));
+	if (rec == NULL) {
 		return;
 	}
-	pthread_mutex_lock (&logsys_cond_mutex);
-	pthread_cond_signal (&logsys_cond);
-	pthread_mutex_unlock (&logsys_cond_mutex);
+	
+	length = strlen (buffer);
+
+	rec->rec_ident = rec_ident;
+	rec->file_name = file_name;
+	rec->function_name = function_name;
+	rec->file_line = file_line;
+	rec->buffer = malloc (length + 1);
+	if (rec->buffer == NULL) {
+		free (rec);
+		return;
+	}
+	memcpy (rec->buffer, buffer, length + 1);
+	
+	list_init (&rec->list);
+	logsys_wthread_lock();
+	logsys_memory_used += length + 1 + sizeof (struct record);
+	if (logsys_memory_used > 512000) {
+		free (rec->buffer);
+		free (rec);
+		logsys_memory_used = logsys_memory_used - length - 1 - sizeof (struct record);
+		logsys_dropped_messages += 1;
+		logsys_wthread_unlock();
+		return;
+		
+	} else {
+		list_add_tail (&rec->list, &logsys_print_finished_records);
+	}
+	logsys_wthread_unlock();
+
+	sem_post (&logsys_print_finished);
 }
 
-static inline void wthread_wait (void)
-{
-	pthread_mutex_lock (&logsys_cond_mutex);
-	pthread_cond_wait (&logsys_cond, &logsys_cond_mutex);
-	pthread_mutex_unlock (&logsys_cond_mutex);
-}
-
-static inline void wthread_wait_locked (void)
-{
-	pthread_cond_wait (&logsys_cond, &logsys_cond_mutex);
-	pthread_mutex_unlock (&logsys_cond_mutex);
-}
-
-static void *logsys_worker_thread (void *data) __attribute__((__noreturn__));
+static void *logsys_worker_thread (void *data) __attribute__((noreturn));
 static void *logsys_worker_thread (void *data)
 {
-	int log_msg;
-	char buf[COMBINE_BUFFER_SIZE];
+	struct record *rec;
+	int dropped = 0;
+	int res;
 
 	/*
 	 * Signal wthread_create that the initialization process may continue
 	 */
-	wthread_signal ();
-	logsys_lock();
-	log_rec_idx = flt_data[FDTAIL_INDEX];
-	logsys_unlock();
-
+	sem_post (&logsys_thread_start);
 	for (;;) {
-		wthread_wait ();
-		/*
-		 * Read and copy the logging record index position
-		 * It may have been updated by records_reclaim if
-		 * messages were lost or or log_rec on the first new
-		 * logging record available
-		 */
-		/*
-		 * Process any pending log messages here
-		 */
-		for (;;) {
-			int yield_counter = 1;
-
-			logsys_lock();
-			if (log_requests_lost > 0) {
-				printf ("lost %d log requests\n", log_requests_lost);
-				log_requests_pending -= log_requests_lost;
-				log_requests_lost = 0;
-			}
-			if (log_requests_pending == 0) {
-				logsys_unlock();
-				break;
-			}
-			log_rec_idx = record_read (buf, log_rec_idx, &log_msg);
-			if (log_msg) {
-				log_requests_pending -= 1;
-			}
-			logsys_unlock();
-
+		dropped = 0;
+retry_sem_wait:
+		res = sem_wait (&logsys_print_finished);
+		if (res == -1 && errno == EINTR) {
+			goto retry_sem_wait;
+		} else
+		if (res == -1) {
 			/*
-			 * print the stored buffer
-			 */
-			if (log_msg) {
-				record_print (buf);
-				if (yield_counter++ > logsys_after_log_ops_yield) {
-					yield_counter = 0;
-					sched_yield ();
-				}
-			}
-		}
-
-		if (wthread_should_exit) {
+  			 * This case shouldn't happen
+  			 */
 			pthread_exit (NULL);
 		}
+		
+
+		logsys_wthread_lock();
+		if (wthread_should_exit) {
+			int value;
+
+			res = sem_getvalue (&logsys_print_finished, &value);
+			if (value == 0) {
+				logsys_wthread_unlock();
+				pthread_exit (NULL);
+			}
+		}
+
+		rec = list_entry (logsys_print_finished_records.next, struct record, list);
+		list_del (&rec->list);
+		logsys_memory_used = logsys_memory_used - strlen (rec->buffer) -
+			sizeof (struct record) - 1;
+		dropped = logsys_dropped_messages;
+		logsys_dropped_messages = 0;
+		logsys_wthread_unlock();
+		if (dropped) {
+			printf ("%d messages lost\n", dropped);
+		}
+		log_printf_to_logs (
+			rec->rec_ident,
+			rec->file_name,
+			rec->function_name,
+			rec->file_line,
+			rec->buffer);
+		free (rec->buffer);
+		free (rec);
 	}
 }
 
@@ -797,29 +765,24 @@ static void wthread_create (void)
 
 	wthread_active = 1;
 
-	pthread_mutex_init (&logsys_cond_mutex, NULL);
-	pthread_cond_init (&logsys_cond, NULL);
-	pthread_mutex_lock (&logsys_cond_mutex);
 
 	/*
 	 * TODO: propagate pthread_create errors back to the caller
 	 */
 	res = pthread_create (&logsys_thread_id, NULL,
 		logsys_worker_thread, NULL);
+	sem_wait (&logsys_thread_start);
 
 	if (res == 0) {
-		/*
-		 * Wait for thread to be started
-		 */
-		wthread_wait_locked ();
 		if (logsys_sched_param_queued == 1) {
 			/*
 			 * TODO: propagate logsys_thread_priority_set errors back to
 			 * the caller
 			 */
-			res = logsys_thread_priority_set (logsys_sched_policy,
-						    &logsys_sched_param,
-						    logsys_after_log_ops_yield);
+			res = logsys_thread_priority_set (
+				logsys_sched_policy,
+				&logsys_sched_param,
+				logsys_after_log_ops_yield);
 			logsys_sched_param_queued = 0;
 		}
 	} else {
@@ -1039,7 +1002,7 @@ int _logsys_system_setup(
 	return (0);
 }
 
-unsigned int _logsys_subsys_create (const char *subsys)
+int _logsys_subsys_create (const char *subsys)
 {
 	int i;
 
@@ -1082,14 +1045,17 @@ int _logsys_wthread_create (void)
 
 int _logsys_rec_init (unsigned int fltsize)
 {
-	/*
-	 * we need to allocate:
-	 * - requested size +
-	 *   2 extra unsigned ints for HEAD/TAIL tracking
-	 *
-	 * then round it up to the next PAGESIZE
-	 */
 	size_t flt_real_size;
+	int res;
+
+	sem_init (&logsys_thread_start, 0, 0);
+
+	sem_init (&logsys_print_finished, 0, 0);
+
+#if defined(HAVE_PTHREAD_SPIN_LOCK)
+	pthread_spin_init (&logsys_flt_spinlock, 1);
+	pthread_spin_init (&logsys_wthread_spinlock, 1);
+#endif
 
 	/*
 	 * XXX: kill me for 1.1 because I am a dirty hack
@@ -1102,34 +1068,32 @@ int _logsys_rec_init (unsigned int fltsize)
 		fltsize = 64000;
 	}
 
-	flt_real_size = ROUNDUP(
-			(fltsize + (2 * sizeof (unsigned int))),
-			sysconf(_SC_PAGESIZE));
+	flt_real_size = ROUNDUP(fltsize, sysconf(_SC_PAGESIZE)) * 4;
 
-	flt_data = malloc (flt_real_size);
-	if (flt_data == NULL) {
+	res = circular_memory_map ((void **)&flt_data, flt_real_size);
+	if (res == -1) {
+		sem_destroy (&logsys_thread_start);
+		sem_destroy (&logsys_print_finished);
+#if defined(HAVE_PTHREAD_SPIN_LOCK)
+		pthread_spin_destroy (&logsys_flt_spinlock);
+		pthread_spin_destroy (&logsys_wthread_spinlock);
+#endif
 		return (-1);
 	}
 
+	memset (flt_data, 0, flt_real_size * 2);
 	/*
 	 * flt_data_size tracks data by ints and not bytes/chars.
-	 *
-	 * the last 2 ints are reserved to store HEAD/TAIL information.
-	 * hide them from the rotating buffer.
 	 */
 
-	flt_data_size = ((flt_real_size / sizeof (unsigned int)) - 2);
+	flt_data_size = flt_real_size / sizeof (uint32_t);
 
 	/*
 	 * First record starts at zero
 	 * Last record ends at zero
 	 */
-	flt_data[FDHEAD_INDEX] = 0;
-	flt_data[FDTAIL_INDEX] = 0;
-
-#if defined(HAVE_PTHREAD_SPIN_LOCK)
-	pthread_spin_init (&logsys_idx_spinlock, 0);
-#endif
+	flt_head = 0;
+	flt_tail = 0;
 
 	return (0);
 }
@@ -1203,7 +1167,8 @@ void _logsys_log_rec (
 		record_reclaim_size += ((buf_len[i] + 3) >> 2) + 1;
 	}
 
-	idx = flt_data[FDHEAD_INDEX];
+	logsys_flt_lock();
+	idx = flt_head;
 	index_start = idx;
 
 	/*
@@ -1230,78 +1195,32 @@ void _logsys_log_rec (
 	 */
 	for (i = 0; i < arguments; i++) {
 		unsigned int bytes;
-		unsigned int full_words;
 		unsigned int total_words;
 
 		bytes = buf_len[i];
-		full_words = bytes >> 2;
 		total_words = (bytes + 3) >> 2;
 
 		flt_data[idx++] = total_words;
 		idx_word_step(idx);
 
-		/*
-		 * determine if this is a wrapped write or normal write
-		 */
-		if (idx + total_words < flt_data_size) {
-			/*
-			 * dont need to wrap buffer
-			 */
-			my_memcpy_32bit (&flt_data[idx], buf_args[i], full_words);
-			if (bytes % 4) {
-				my_memcpy_8bit ((char *)&flt_data[idx + full_words],
-					((const char *)buf_args[i]) + (full_words << 2), bytes % 4);
-			}
-		} else {
-			/*
-			 * need to wrap buffer
-			 */
-			unsigned int first;
-			unsigned int second;
+		memcpy (&flt_data[idx], buf_args[i], buf_len[i]);
 
-			first = flt_data_size - idx;
-			if (first > full_words) {
-				first = full_words;
-			}
-			second = full_words - first;
-			my_memcpy_32bit (&flt_data[idx],
-					 (const int *)buf_args[i], first);
-			my_memcpy_32bit (&flt_data[0],
-				(const int *)(((const unsigned char *)buf_args[i]) + (first << 2)),
-				second);
-			if (bytes % 4) {
-				my_memcpy_8bit ((char *)&flt_data[0 + second],
-					((const char *)buf_args[i]) + (full_words << 2), bytes % 4);
-			}
-		}
 		idx += total_words;
 		idx_buffer_step (idx);
+
 	}
 	words_written = idx - index_start;
 	if (words_written < 0) {
 		words_written += flt_data_size;
 	}
-
 	/*
 	 * Commit the write of the record size now that the full record
 	 * is in the memory buffer
 	 */
 	flt_data[index_start] = words_written;
 
-	/*
-	 * If the index of the current head equals the current log_rec_idx,
-	 * and this is not a log_printf operation, set the log_rec_idx to
-	 * the new head position and commit the new head.
-	 */
-	logsys_lock();
-	if (LOGSYS_DECODE_RECID(rec_ident) == LOGSYS_RECID_LOG) {
-		log_requests_pending += 1;
-	}
-	if (log_requests_pending == 0) {
-		log_rec_idx = idx;
-	}
-	flt_data[FDHEAD_INDEX] = idx;
-	logsys_unlock();
+	flt_head = idx;
+	logsys_flt_unlock();
 	records_written++;
 }
 
@@ -1322,12 +1241,6 @@ void _logsys_log_vprintf (
 	subsysid = LOGSYS_DECODE_SUBSYSID(rec_ident);
 	level = LOGSYS_DECODE_LEVEL(rec_ident);
 
-	if ((level > logsys_loggers[subsysid].syslog_priority) &&
-	    (level > logsys_loggers[subsysid].logfile_priority) &&
-	    (logsys_loggers[subsysid].debug == 0)) {
-		return;
-	}
-
 	len = vsprintf (logsys_print_buffer, format, ap);
 	if (logsys_print_buffer[len - 1] == '\n') {
 		logsys_print_buffer[len - 1] = '\0';
@@ -1354,19 +1267,39 @@ void _logsys_log_vprintf (
 		logsys_print_buffer, len + 1,
 		LOGSYS_REC_END);
 
+	/*
+	 * If logsys is not going to print a message to a log target don't
+	 * queue one
+	 */
+	if ((level > logsys_loggers[subsysid].syslog_priority && 
+		level > logsys_loggers[subsysid].logfile_priority &&
+			logsys_loggers[subsysid].debug == 0) ||
+
+		(level == LOGSYS_LEVEL_DEBUG &&
+			logsys_loggers[subsysid].debug == 0)) {
+
+		return;
+	}
+
 	if ((logsys_loggers[LOGSYS_MAX_SUBSYS_COUNT].mode & LOGSYS_MODE_THREADED) == 0) {
 		/*
 		 * Output (and block) if the log mode is not threaded otherwise
 		 * expect the worker thread to output the log data once signaled
 		 */
 		log_printf_to_logs (rec_ident,
-				    short_file_name, function_name, file_line,
-				    logsys_print_buffer);
+			short_file_name,
+			function_name,
+			file_line,
+			logsys_print_buffer);
 	} else {
 		/*
 		 * Signal worker thread to display logging output
 		 */
-		wthread_signal ();
+		log_printf_to_logs_wthread (rec_ident,
+			short_file_name,
+			function_name,
+			file_line,
+			logsys_print_buffer);
 	}
 }
 
@@ -1408,7 +1341,7 @@ void logsys_fork_completed (void)
 	_logsys_wthread_create ();
 }
 
-unsigned int logsys_config_mode_set (const char *subsys, unsigned int mode)
+int logsys_config_mode_set (const char *subsys, unsigned int mode)
 {
 	int i;
 
@@ -1497,7 +1430,7 @@ char *logsys_format_get (void)
 	return format_buffer;
 }
 
-unsigned int logsys_config_syslog_facility_set (
+int logsys_config_syslog_facility_set (
 	const char *subsys,
 	unsigned int facility)
 {
@@ -1525,7 +1458,7 @@ unsigned int logsys_config_syslog_facility_set (
 	return i;
 }
 
-unsigned int logsys_config_syslog_priority_set (
+int logsys_config_syslog_priority_set (
 	const char *subsys,
 	unsigned int priority)
 {
@@ -1549,7 +1482,7 @@ unsigned int logsys_config_syslog_priority_set (
 	return i;
 }
 
-unsigned int logsys_config_logfile_priority_set (
+int logsys_config_logfile_priority_set (
 	const char *subsys,
 	unsigned int priority)
 {
@@ -1573,7 +1506,7 @@ unsigned int logsys_config_logfile_priority_set (
 	return i;
 }
 
-unsigned int logsys_config_debug_set (
+int logsys_config_debug_set (
 	const char *subsys,
 	unsigned int debug)
 {
@@ -1652,11 +1585,14 @@ int logsys_thread_priority_set (
 
 {
 	int res = 0;
+	if (param == NULL) {
+		return (0);
+	}
 
 #if defined(HAVE_PTHREAD_SETSCHEDPARAM) && defined(HAVE_SCHED_GET_PRIORITY_MAX)
 	if (wthread_active == 0) {
 		logsys_sched_policy = policy;
-		memcpy(&logsys_sched_param, &param, sizeof(struct sched_param));
+		memcpy(&logsys_sched_param, param, sizeof(struct sched_param));
 		logsys_sched_param_queued = 1;
 	} else {
 		res = pthread_setschedparam (logsys_thread_id, policy, param);
@@ -1673,42 +1609,85 @@ int logsys_thread_priority_set (
 int logsys_log_rec_store (const char *filename)
 {
 	int fd;
-	ssize_t written_size;
-	size_t size_to_write = (flt_data_size + 2) * sizeof (unsigned int);
+	ssize_t written_size = 0;
+	size_t this_write_size;
 
 	fd = open (filename, O_CREAT|O_RDWR, 0700);
 	if (fd < 0) {
 		return (-1);
 	}
 
-	written_size = write (fd, &flt_data_size, sizeof(unsigned int));
-	if ((written_size < 0) || (written_size != sizeof(unsigned int))) {
-		close (fd);
-		return (-1);
+	this_write_size = write (fd, &flt_data_size, sizeof(uint32_t));
+	if (this_write_size != sizeof(unsigned int)) {
+		goto error_exit;
 	}
+	written_size += this_write_size;
 
-	written_size = write (fd, flt_data, size_to_write);
-	if (close (fd) != 0)
-		return (-1);
-	if (written_size < 0) {
-		return (-1);
-	} else if ((size_t)written_size != size_to_write) {
-		return (-1);
+	this_write_size = write (fd, flt_data, flt_data_size * sizeof (uint32_t));
+	if (this_write_size != (flt_data_size * sizeof(uint32_t))) {
+		goto error_exit;
 	}
+	written_size += this_write_size;
 
+	this_write_size = write (fd, &flt_head, sizeof (uint32_t));
+	if (this_write_size != (sizeof(uint32_t))) {
+		goto error_exit;
+	}
+	written_size += this_write_size;
+	this_write_size = write (fd, &flt_tail, sizeof (uint32_t));
+	if (this_write_size != (sizeof(uint32_t))) {
+		goto error_exit;
+	}
+	written_size += this_write_size;
+	if (written_size != ((flt_data_size + 3) * sizeof (uint32_t))) { 
+		goto error_exit;
+	}
+	
 	return (0);
+
+error_exit:
+	close (fd);
+	return (-1);
 }
 
 void logsys_atexit (void)
 {
-	if (wthread_active) {
+	int res;
+	int value;
+	struct record *rec;
+
+	if (wthread_active == 0) {
+		for (;;) {
+			logsys_wthread_lock();
+
+			res = sem_getvalue (&logsys_print_finished, &value);
+			if (value == 0) {
+				logsys_wthread_unlock();
+				return;
+			}
+			sem_wait (&logsys_print_finished);
+
+			rec = list_entry (logsys_print_finished_records.next, struct record, list);
+			list_del (&rec->list);
+			logsys_memory_used = logsys_memory_used - strlen (rec->buffer) -
+				sizeof (struct record) - 1;
+			logsys_wthread_unlock();
+			log_printf_to_logs (
+				rec->rec_ident,
+				rec->file_name,
+				rec->function_name,
+				rec->file_line,
+				rec->buffer);
+			free (rec->buffer);
+			free (rec);
+		}
+	} else {
 		wthread_should_exit = 1;
-		wthread_signal ();
+		sem_post (&logsys_print_finished);
 		pthread_join (logsys_thread_id, NULL);
 	}
 }
 
 void logsys_flush (void)
 {
-	wthread_signal ();
 }
