@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2006 MontaVista Software, Inc.
- * Copyright (c) 2007-2009 Red Hat, Inc.
+ * Copyright (c) 2007-2010 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -33,11 +33,15 @@
  * THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#define _XOPEN_SOURCE 600
+
 #include <config.h>
 
 #include <stdio.h>
 #include <errno.h>
 
+#include <pthread.h>
+
 #include <corosync/list.h>
 #include <corosync/hdb.h>
 #include <corosync/lcr/lcr_comp.h>
@@ -97,44 +101,21 @@ struct object_find_instance {
 
 struct objdb_iface_ver0 objdb_iface;
 struct list_head objdb_trackers_head;
-static pthread_rwlock_t reload_lock;
-static pthread_t lock_thread;
-static pthread_mutex_t meta_lock;
+static pthread_mutex_t objdb_mutex;
+static pthread_mutexattr_t objdb_mutex_attr;
 
 DECLARE_HDB_DATABASE (object_instance_database,NULL);
 
 DECLARE_HDB_DATABASE (object_find_instance_database,NULL);
 
-static void objdb_wrlock(void)
+static void objdb_lock(void)
 {
-	pthread_mutex_lock(&meta_lock);
-	pthread_rwlock_wrlock(&reload_lock);
-	lock_thread = pthread_self();
-	pthread_mutex_unlock(&meta_lock);
+	pthread_mutex_lock(&objdb_mutex);
 }
 
-static void objdb_rdlock(void)
+static void objdb_unlock(void)
 {
-	pthread_mutex_lock(&meta_lock);
-	if (lock_thread != pthread_self())
-		pthread_rwlock_rdlock(&reload_lock);
-	pthread_mutex_unlock(&meta_lock);
-}
-
-static void objdb_rdunlock(void)
-{
-	pthread_mutex_lock(&meta_lock);
-	if (lock_thread != pthread_self())
-		pthread_rwlock_unlock(&reload_lock);
-	pthread_mutex_unlock(&meta_lock);
-}
-
-static void objdb_wrunlock(void)
-{
-	pthread_mutex_lock(&meta_lock);
-	pthread_rwlock_unlock(&reload_lock);
-	lock_thread = 0;
-	pthread_mutex_unlock(&meta_lock);
+	pthread_mutex_unlock(&objdb_mutex);
 }
 
 static int objdb_init (void)
@@ -166,8 +147,11 @@ static int objdb_init (void)
 	list_init (&instance->child_list);
 	list_init (&instance->track_head);
 	list_init (&objdb_trackers_head);
-	pthread_rwlock_init(&reload_lock, NULL);
-	pthread_mutex_init(&meta_lock, NULL);
+
+	pthread_mutexattr_init(&objdb_mutex_attr);
+
+	pthread_mutexattr_settype(&objdb_mutex_attr, PTHREAD_MUTEX_RECURSIVE);
+	pthread_mutex_init(&objdb_mutex, &objdb_mutex_attr);
 
 	hdb_handle_put (&object_instance_database, handle);
 	return (0);
@@ -352,6 +336,7 @@ static void object_reload_notification(int startstop, int flush)
 			if (tmptracker_pt) {
 				list_add(&tmptracker_pt->object_list, &tmplist);
 				tmptracker_pt->object_reload_notify_fn = tracker_pt->object_reload_notify_fn;
+				tmptracker_pt->data_pt = tracker_pt->data_pt;
 			}
 		}
 	}
@@ -384,7 +369,7 @@ static int object_create (
 	int found = 0;
 	int i;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		parent_object_handle, (void *)&parent_instance);
 	if (res != 0) {
@@ -457,7 +442,7 @@ static int object_create (
 		object_instance->parent_handle,
 		object_instance->object_name,
 		object_instance->object_name_len);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_put_destroy:
@@ -470,7 +455,7 @@ error_object_put:
 	hdb_handle_put (&object_instance_database, parent_object_handle);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -481,7 +466,7 @@ static int object_priv_set (
 	int res;
 	struct object_instance *object_instance;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&object_instance);
@@ -492,11 +477,11 @@ static int object_priv_set (
 	object_instance->priv = priv;
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -517,7 +502,7 @@ static int object_key_create_typed(
 	size_t expected_size;
 	int test_size_by_type = CS_TRUE;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 			  object_handle, (void *)&instance);
@@ -637,7 +622,7 @@ static int object_key_create_typed(
 	object_key_changed_notification(object_handle, key_name, key_len,
 		value, value_len, OBJECT_KEY_CREATED);
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_put_key:
@@ -650,7 +635,7 @@ error_put:
 	hdb_handle_put (&object_instance_database, object_handle);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -726,12 +711,12 @@ static int object_destroy (
 	struct object_instance *instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
-		objdb_rdunlock();
+		objdb_unlock();
 		return (res);
 	}
 
@@ -748,7 +733,7 @@ static int object_destroy (
 	hdb_handle_put (&object_instance_database, object_handle);
 	hdb_handle_destroy (&object_instance_database, object_handle);
 
-	objdb_rdunlock();
+	objdb_unlock();
 	return (res);
 }
 
@@ -760,7 +745,7 @@ static int object_valid_set (
 	struct object_instance *instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
@@ -772,11 +757,11 @@ static int object_valid_set (
 
 	hdb_handle_put (&object_instance_database, object_handle);
 
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -788,7 +773,7 @@ static int object_key_valid_set (
 	struct object_instance *instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
@@ -800,11 +785,11 @@ static int object_key_valid_set (
 
 	hdb_handle_put (&object_instance_database, object_handle);
 
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -821,7 +806,7 @@ static int object_find_create (
 	struct object_instance *object_instance;
 	struct object_find_instance *object_find_instance;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&object_instance);
 	if (res != 0) {
@@ -847,7 +832,7 @@ static int object_find_create (
 	hdb_handle_put (&object_instance_database, object_handle);
 	hdb_handle_put (&object_find_instance_database, *object_find_handle);
 
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_destroy:
@@ -857,7 +842,7 @@ error_put:
 	hdb_handle_put (&object_instance_database, object_handle);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -871,7 +856,7 @@ static int object_find_next (
 	struct list_head *list;
 	unsigned int found = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_find_instance_database,
 		object_find_handle, (void *)&object_find_instance);
 	if (res != 0) {
@@ -902,11 +887,11 @@ static int object_find_next (
 		*object_handle = object_instance->object_handle;
 		res = 0;
 	}
-	objdb_rdunlock();
+	objdb_unlock();
 	return (res);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -916,7 +901,7 @@ static int object_find_destroy (
 	struct object_find_instance *object_find_instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_find_instance_database,
 		object_find_handle, (void *)&object_find_instance);
 	if (res != 0) {
@@ -925,11 +910,11 @@ static int object_find_destroy (
 	hdb_handle_put(&object_find_instance_database, object_find_handle);
 	hdb_handle_destroy(&object_find_instance_database, object_find_handle);
 
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -947,7 +932,7 @@ static int object_key_get_typed (
 	int found = 0;
 	size_t key_len = strlen(key_name);
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
@@ -976,11 +961,11 @@ static int object_key_get_typed (
 	}
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (res);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1024,7 +1009,7 @@ static int object_key_increment (
 	struct list_head *list;
 	int found = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
@@ -1086,11 +1071,15 @@ static int object_key_increment (
 	}
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	if (res == 0) {
+		object_key_changed_notification (object_handle, key_name, key_len,
+			object_key->value, object_key->value_len, OBJECT_KEY_REPLACED);
+	}
+	objdb_unlock();
 	return (res);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1106,7 +1095,7 @@ static int object_key_decrement (
 	struct list_head *list;
 	int found = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
@@ -1169,11 +1158,15 @@ static int object_key_decrement (
 	}
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	if (res == 0) {
+		object_key_changed_notification (object_handle, key_name, key_len,
+			object_key->value, object_key->value_len, OBJECT_KEY_REPLACED);
+	}
+	objdb_unlock();
 	return (res);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1189,7 +1182,7 @@ static int object_key_delete (
 	struct list_head *list;
 	int found = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
 	if (res != 0) {
@@ -1222,11 +1215,11 @@ static int object_key_delete (
 		object_key_changed_notification(object_handle, key_name, key_len,
 			NULL, 0, OBJECT_KEY_DELETED);
 	}
-	objdb_rdunlock();
+	objdb_unlock();
 	return (ret);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1243,8 +1236,9 @@ static int object_key_replace (
 	struct object_key *object_key = NULL;
 	struct list_head *list;
 	int found = 0;
+	int value_changed = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&instance);
@@ -1306,9 +1300,17 @@ static int object_key_replace (
 				goto error_exit;
 			free(object_key->value);
 			object_key->value = replacement_value;
+			memset (object_key->value, 0, new_value_len);
+			object_key->value_len = new_value_len;
+		}
+		if (memcmp (object_key->value, new_value, new_value_len) == 0) {
+			value_changed = 0;
+		}
+		else {
+			memcpy(object_key->value, new_value, new_value_len);
+			object_key->value_len = new_value_len;
+			value_changed = 1;
 		}
-		memcpy(object_key->value, new_value, new_value_len);
-		object_key->value_len = new_value_len;
 	}
 	else {
 		ret = -1;
@@ -1316,17 +1318,17 @@ static int object_key_replace (
 	}
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	if (ret == 0) {
+	if (ret == 0 && value_changed) {
 		object_key_changed_notification (object_handle, key_name, key_len,
 			new_value, new_value_len, OBJECT_KEY_REPLACED);
 	}
-	objdb_rdunlock();
+	objdb_unlock();
 	return (ret);
 
 error_put:
 	hdb_handle_put (&object_instance_database, object_handle);
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1337,7 +1339,7 @@ static int object_priv_get (
 	int res;
 	struct object_instance *object_instance;
 
-	objdb_rdunlock();
+	objdb_unlock();
 	res = hdb_handle_get (&object_instance_database,
 		object_handle, (void *)&object_instance);
 	if (res != 0) {
@@ -1347,11 +1349,11 @@ static int object_priv_get (
 	*priv = object_instance->priv;
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1445,7 +1447,7 @@ static int object_key_iter_reset(hdb_handle_t object_handle)
 	unsigned int res;
 	struct object_instance *instance;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 			      object_handle, (void *)&instance);
@@ -1455,11 +1457,11 @@ static int object_key_iter_reset(hdb_handle_t object_handle)
 	instance->iter_key_list = &instance->key_head;
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (0);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1475,7 +1477,7 @@ static int object_key_iter_typed (hdb_handle_t parent_object_handle,
 	struct list_head *list;
 	unsigned int found = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 		parent_object_handle, (void *)&instance);
@@ -1502,11 +1504,11 @@ static int object_key_iter_typed (hdb_handle_t parent_object_handle,
 	}
 
 	hdb_handle_put (&object_instance_database, parent_object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (res);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1521,8 +1523,10 @@ static int object_key_iter(hdb_handle_t parent_object_handle,
 	char *str;
 	ret = object_key_iter_typed (parent_object_handle,
 		(char**)key_name, value, value_len, &t);
-	str = *key_name;
-	*key_len = strlen(str);
+	if (!ret) {
+		str = *key_name;
+		*key_len = strlen(str);
+	}
 	return ret;
 }
 
@@ -1540,7 +1544,7 @@ static int object_key_iter_from(hdb_handle_t parent_object_handle,
 	struct list_head *list;
 	unsigned int found = 0;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 		parent_object_handle, (void *)&instance);
@@ -1574,11 +1578,11 @@ static int object_key_iter_from(hdb_handle_t parent_object_handle,
 	}
 
 	hdb_handle_put (&object_instance_database, parent_object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 	return (res);
 
 error_exit:
-	objdb_rdunlock();
+	objdb_unlock();
 	return (-1);
 }
 
@@ -1589,12 +1593,12 @@ static int object_parent_get(hdb_handle_t object_handle,
 	struct object_instance *instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 
 	res = hdb_handle_get (&object_instance_database,
 			      object_handle, (void *)&instance);
 	if (res != 0) {
-		objdb_rdunlock();
+		objdb_unlock();
 		return (res);
 	}
 
@@ -1604,7 +1608,7 @@ static int object_parent_get(hdb_handle_t object_handle,
 		*parent_handle = instance->parent_handle;
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 
 	return (0);
 }
@@ -1616,11 +1620,11 @@ static int object_name_get(hdb_handle_t object_handle,
 	struct object_instance *instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 			      object_handle, (void *)&instance);
 	if (res != 0) {
-		objdb_rdunlock();
+		objdb_unlock();
 		return (res);
 	}
 
@@ -1628,7 +1632,7 @@ static int object_name_get(hdb_handle_t object_handle,
 	*object_name_len = instance->object_name_len;
 
 	hdb_handle_put (&object_instance_database, object_handle);
-	objdb_rdunlock();
+	objdb_unlock();
 
 	return (0);
 }
@@ -1727,11 +1731,11 @@ static int object_dump(hdb_handle_t object_handle,
 	struct object_instance *instance;
 	unsigned int res;
 
-	objdb_rdlock();
+	objdb_lock();
 	res = hdb_handle_get (&object_instance_database,
 			      object_handle, (void *)&instance);
 	if (res != 0) {
-		objdb_rdunlock();
+		objdb_unlock();
 		return (res);
 	}
 
@@ -1739,7 +1743,7 @@ static int object_dump(hdb_handle_t object_handle,
 
 	hdb_handle_put (&object_instance_database, object_handle);
 
-	objdb_rdunlock();
+	objdb_unlock();
 	return (res);
 }
 
@@ -1752,18 +1756,18 @@ static int object_write_config(const char **error_string)
 
 	main_get_config_modules(&modules, &num_modules);
 
-	objdb_wrlock();
+	objdb_lock();
 
 	for (i=0; i<num_modules; i++) {
 		if (modules[i]->config_writeconfig) {
 			res = modules[i]->config_writeconfig(&objdb_iface, error_string);
 			if (res) {
-				objdb_wrunlock();
+				objdb_unlock();
 				return res;
 			}
 		}
 	}
-	objdb_wrunlock();
+	objdb_unlock();
 	return 0;
 }
 
@@ -1777,19 +1781,19 @@ static int object_reload_config(int flush, const char **error_string)
 	main_get_config_modules(&modules, &num_modules);
 	object_reload_notification(OBJDB_RELOAD_NOTIFY_START, flush);
 
-	objdb_wrlock();
+	objdb_lock();
 
 	for (i=0; i<num_modules; i++) {
 		if (modules[i]->config_reloadconfig) {
 			res = modules[i]->config_reloadconfig(&objdb_iface, flush, error_string);
 			if (res) {
 				object_reload_notification(OBJDB_RELOAD_NOTIFY_FAILED, flush);
-				objdb_wrunlock();
+				objdb_unlock();
 				return res;
 			}
 		}
 	}
-	objdb_wrunlock();
+	objdb_unlock();
 	object_reload_notification(OBJDB_RELOAD_NOTIFY_END, flush);
 	return 0;
 }
