@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2002-2005 MontaVista Software, Inc.
- * Copyright (c) 2006-2009 Red Hat, Inc.
+ * Copyright (c) 2006-2010 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -421,7 +421,12 @@ int totem_config_validate (
 		/*
 		 * Some error checking of parsed data to make sure its valid
 		 */
-		if ((int *)&totem_config->interfaces[i].mcast_addr.addr == 0) {
+
+		struct totem_ip_address null_addr;
+		memset (&null_addr, 0, sizeof (struct totem_ip_address));
+
+		if (memcmp (&totem_config->interfaces[i].mcast_addr, &null_addr,
+				sizeof (struct totem_ip_address)) == 0) {
 			error_reason = "No multicast address specified";
 			goto parse_error;
 		}
@@ -559,13 +564,6 @@ int totem_config_validate (
 		goto parse_error;
 	}
 
-	if (totem_config->consensus_timeout < 1.2 * totem_config->token_timeout) {
-		snprintf (local_error_reason, sizeof(local_error_reason),
-			"The consensus timeout parameter (%d ms) must be atleast 1.2 * token (%d ms).",
-			totem_config->consensus_timeout, (int) ((float)1.2 * totem_config->token_timeout));
-		goto parse_error;
-	}
-
 	if (totem_config->merge_timeout == 0) {
 		totem_config->merge_timeout = MERGE_TIMEOUT;
 	}
@@ -790,8 +788,8 @@ int totem_config_keyread (
 
 	return (0);
 
-	*error_string = error_string_response;
 key_error:
+	*error_string = error_string_response;
 	return (-1);
 
 }
@@ -818,6 +816,9 @@ static void totem_objdb_reload_notify(objdb_reload_notify_type_t type, int flush
 	struct totem_config *totem_config = priv_data_pt;
 	hdb_handle_t totem_object_handle;
 
+	if (totem_config == NULL)
+	        return;
+
 	/*
 	 * A new totem {} key might exist, cancel the
 	 * existing notification at the start of reload,
@@ -831,7 +832,7 @@ static void totem_objdb_reload_notify(objdb_reload_notify_type_t type, int flush
 			NULL,
 			NULL,
 			NULL,
-			NULL);
+			totem_config);
 	}
 
 	if (type == OBJDB_RELOAD_NOTIFY_END ||
@@ -840,8 +841,14 @@ static void totem_objdb_reload_notify(objdb_reload_notify_type_t type, int flush
 
 		if (!totem_handle_find(global_objdb,
 				      &totem_object_handle)) {
-			add_totem_config_notification(global_objdb, totem_config, totem_object_handle);
 
+		        global_objdb->object_track_start(totem_object_handle,
+						  1,
+						  totem_key_change_notify,
+						  NULL, // object_create_notify,
+						  NULL, // object_destroy_notify,
+						  NULL, // object_reload_notify
+						  totem_config); // priv_data
 			/*
 			 * Reload the configuration
 			 */
