@@ -68,6 +68,8 @@
 #include <corosync/ipc_votequorum.h>
 #include <corosync/list.h>
 
+#include "../exec/tlist.h"
+
 #define VOTEQUORUM_MAJOR_VERSION 7
 #define VOTEQUORUM_MINOR_VERSION 0
 #define VOTEQUORUM_PATCH_VERSION 0
@@ -107,7 +109,7 @@ struct cluster_node {
 
 	nodestate_t state;
 
-	struct timeval last_hello; /* Only used for quorum devices */
+	unsigned long long int last_hello; /* Only used for quorum devices */
 
 	struct list_head list;
 };
@@ -533,6 +535,9 @@ static int votequorum_exec_init_fn (struct corosync_api_v1 *api)
 	add_votequorum_config_notification(object_handle);
 	corosync_api->object_find_destroy(find_handle);
 
+	/* Start us off with one node */
+	quorum_exec_send_nodeinfo();
+
 	LEAVE();
 	return (0);
 }
@@ -935,7 +940,6 @@ static void quorum_confchg_fn (
 				node->flags |= NODE_FLAGS_BEENDOWN;
 			}
 		}
-		recalculate_quorum(leaving, leaving);
 	}
 
 	if (member_list_entries) {
@@ -947,6 +951,9 @@ static void quorum_confchg_fn (
 		quorum_exec_send_nodeinfo();
 	}
 
+	if (left_list_entries)
+		recalculate_quorum(leaving, leaving);
+
 	memcpy(&quorum_ringid, ring_id, sizeof(*ring_id));
 	LEAVE();
 }
@@ -1042,7 +1049,8 @@ static void message_handler_req_exec_votequorum_nodeinfo (
 	}
 	node->flags &= ~NODE_FLAGS_BEENDOWN;
 
-	if (new_node || old_votes != node->votes || old_expected != node->expected_votes || old_state != node->state)
+	if (new_node || req_exec_quorum_nodeinfo->first_trans || 
+	    old_votes != node->votes || old_expected != node->expected_votes || old_state != node->state)
 		recalculate_quorum(0, 0);
 	LEAVE();
 }
@@ -1307,13 +1315,13 @@ static void message_handler_req_lib_votequorum_leaving (void *conn, const void *
 
 static void quorum_device_timer_fn(void *arg)
 {
-	struct timeval now;
-
 	ENTER();
 	if (!quorum_device || quorum_device->state == NODESTATE_DEAD)
 		return;
-	gettimeofday(&now, NULL);
-	if (quorum_device->last_hello.tv_sec + quorumdev_poll/1000 < now.tv_sec) {
+
+	if ( (quorum_device->last_hello / TIMERLIST_NS_IN_SEC) + quorumdev_poll/1000 <
+		(timerlist_nano_current_get () / TIMERLIST_NS_IN_SEC)) {
+
 		quorum_device->state = NODESTATE_DEAD;
 		log_printf(LOGSYS_LEVEL_INFO, "lost contact with quorum device\n");
 		recalculate_quorum(0, 0);
@@ -1395,7 +1403,7 @@ static void message_handler_req_lib_votequorum_qdisk_poll (void *conn,
 
 	if (quorum_device) {
 		if (req_lib_votequorum_qdisk_poll->state) {
-			gettimeofday(&quorum_device->last_hello, NULL);
+			quorum_device->last_hello = timerlist_nano_current_get ();
 			if (quorum_device->state == NODESTATE_DEAD) {
 				quorum_device->state = NODESTATE_MEMBER;
 				recalculate_quorum(0, 0);
